/*
 * JORAM: Java(TM) Open Reliable Asynchronous Messaging
 * Copyright (C) 2001 - 2022 ScalAgent Distributed Technologies
 * Copyright (C) 1996 - 2000 Dyade
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): Frederic Maistre (INRIA)
 * Contributor(s): ScalAgent Distributed Technologies
 */
package org.objectweb.joram.shared.client;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Vector;

import fr.dyade.aaa.common.stream.StreamUtil;

/**
 * A <code>SessDenyRequest</code> instance is used by a <code>Session</code>
 * for denying the messages it consumed.
 */
public final class SessDenyRequest extends AbstractJmsRequest {
  /** define serialVersionUID for interoperability */
  private static final long serialVersionUID = 1L;

  /** Vector of message identifiers. */
  private Vector<String> ids;

  /**
   * Sets the vector of identifiers.
   * 
   * @param ids The identifiers of denied messages.
   */
  public void setIds(Vector<String> ids) {
    this.ids = ids;
  }

  public void addId(String id) {
    if (ids == null) 
      ids = new Vector<String>();
    ids.addElement(id);
  }

  /**
   * Returns the vector of denied messages identifiers.
   * 
   * @return The identifiers of denied messages.
   */
  public Vector<String> getIds() {
    return ids;
  }

  /** <code>true</code> if the request destination is a queue. */
  private boolean queueMode; 

  /**
   * Sets the target destination type.
   * 
   * @param queueMode The destination type.
   */
  public void setQueueMode(boolean queueMode) {
    this.queueMode = queueMode;
  }

  /**
   * Returns <code>true</code> if the request destination is a queue.
   * 
   * @return The destination type.
   */
  public boolean getQueueMode() {
    return queueMode;
  }

  /** <code>true</code> if the request must not be acked by the server. */
  private boolean doNotAck = false;

  /**
   * Sets the server ack policy.
   * 
   * @param doNotAck True if the request doesn't need ack.
   */
  public void setDoNotAck(boolean doNotAck) {
    this.doNotAck = doNotAck;
  }

  /**
   * Returns <code>true</code> if the request must not be acked by the  server.
   * 
   * @return True if the request doesn't need ack.
   */
  public boolean getDoNotAck() {
    return doNotAck;
  }

  /** true if the message has already been delivered.  */
  private boolean redelivered = false;
  
  /**
   * @return the redelivered
   */
  public boolean isRedelivered() {
    return redelivered;
  }

  /**
   * @param redelivered the redelivered to set
   */
  public void setRedelivered(boolean redelivered) {
    this.redelivered = redelivered;
  }

  /**
   * Constructs a <code>SessDenyRequest</code> instance.
   *
   * @param targetName  Name of the target queue or subscription.
   * @param ids  Vector of denied message identifiers.
   * @param queueMode  <code>true</code> if this request is destinated to a queue.
   */
  public SessDenyRequest(String targetName, Vector<String> ids, boolean queueMode) {
    super(targetName);
    this.ids = ids;
    this.queueMode = queueMode;
  }

  /**
   * Constructs a <code>SessDenyRequest</code> instance.
   *
   * @param targetName  Name of the target queue or subscription.
   * @param ids  Vector of denied message identifiers.
   * @param queueMode  <code>true</code> if this request is destinated to a queue.
   * @param doNotAck  <code>true</code> if this request must not be acked by the server.
   */
  public SessDenyRequest(String targetName, Vector<String> ids, boolean queueMode, boolean doNotAck) {
    super(targetName);
    this.ids = ids;
    this.queueMode = queueMode;
    this.doNotAck = doNotAck;
  }

  /**
   * Public no-arg constructor needed by Externalizable.
   */
  public SessDenyRequest() {}

  /* ***** ***** ***** ***** *****
   * Streamable interface
   * ***** ***** ***** ***** ***** */

  protected int getClassId() {
    return SESS_DENY_REQUEST;
  }

  /**
   *  The object implements the writeTo method to write its contents to
   * the output stream.
   *
   * @param os the stream to write the object to
   */
  public void writeTo(OutputStream os) throws IOException {
    super.writeTo(os);
    StreamUtil.writeListOfStringTo(ids, os);
    StreamUtil.writeTo(queueMode, os);
    StreamUtil.writeTo(doNotAck, os);
    StreamUtil.writeTo(redelivered, os);
  }

  /**
   *  The object implements the readFrom method to restore its contents from
   * the input stream.
   *
   * @param is the stream to read data from in order to restore the object
   */
  public void readFrom(InputStream is) throws IOException {
    super.readFrom(is);
    ids = StreamUtil.readVectorOfStringFrom(is);
    queueMode = StreamUtil.readBooleanFrom(is);
    doNotAck = StreamUtil.readBooleanFrom(is);
    redelivered = StreamUtil.readBooleanFrom(is);
  }
}
